#include "stdafx.h"
#include "LexicalAnalyser.h"
#include<ctype.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CToken::CToken(): TokenType(nilT), TokenInt(0)
{

}

CToken::~CToken()
{

}

CLexicalAnalyser::CLexicalAnalyser(): CurrentPosition(NULL)
{

}

CLexicalAnalyser::~CLexicalAnalyser()
{

}

CSymbol::CSymbol(CString string, int type):	SymbolString(string),
											TokenType(type)
{
	
}

CSymbol::~CSymbol()
{

}

bool CSymbol::operator < (CSymbol &Symbol)
{
 return SymbolString.GetLength()< Symbol.SymbolString.GetLength();
}

bool CSymbol::operator <= (CSymbol &Symbol)
{
 return SymbolString.GetLength()<= Symbol.SymbolString.GetLength();
}

bool CSymbol::operator > (CSymbol &Symbol)
{
 return SymbolString.GetLength()> Symbol.SymbolString.GetLength();
}

bool CSymbol::operator >= (CSymbol &Symbol)
{
 return SymbolString.GetLength()>= Symbol.SymbolString.GetLength();
}

void CLexicalAnalyser::setSymbol(CSymbol Symbol)
{
	if(SymbolTable.GetSize()==0)
		SymbolTable.Add(Symbol);
	else
	{
		for(int i=0;i<SymbolTable.GetSize()-1;i++)
		{
			if(SymbolTable[i]>Symbol && SymbolTable[i+1]<=Symbol)
			{
				SymbolTable.InsertAt(i+1,Symbol);
				return;
			}
			else if(SymbolTable[i]<=Symbol)
			{
				SymbolTable.InsertAt(i,Symbol);
				return;
			}
		}
		SymbolTable.Add(Symbol);
	}

}

CSymbol::CSymbol()
{

}

void CLexicalAnalyser::Text2TokenSequence(char* sequence)
{
	CToken token;
	BOOL bInComment = FALSE; 
	BOOL bInMultiComment = FALSE; 

	TokenSequence.RemoveAll();
	CurrentPosition=TokenSequence.GetHeadPosition();

	int seqln = strlen(sequence);

	for(int i=0; i< seqln;i++)
	{
		token.TokenType=nilT;
 
		if(i+1<seqln)
		{
			if(sequence[i]=='-' && sequence[i+1]=='-')
			{
				bInComment = TRUE;
				continue;
			}
			if(sequence[i]=='/' && sequence[i+1]=='*')
			{
				bInMultiComment = TRUE;
				i++;
				continue;
			}
		}

		if(bInComment)
		{
			if(sequence[i]=='\n') bInComment = FALSE;
			continue;
		}

		if(bInMultiComment)
		{
			if(i+1<seqln)
				if(sequence[i]=='*' && sequence[i+1]=='/') 
				{
					bInMultiComment = FALSE;
					i++;
				}
			continue;
		}

		if(sequence[i]==' ' || sequence[i]=='\t')
			continue;
		else if(sequence[i]=='\"') 
		{
			i++;
			token.TokenType=stringT;
			token.TokenName.Empty();

			while(i< seqln && sequence[i]!='\"')
			{
				token.TokenName+=sequence[i];
				i++;
			}
			AddToken(token);
		}
		else if(sequence[i]=='\r' && sequence[i+1]=='\n')
		{
			token.TokenType=endlineT;
			i++;
			AddToken(token);
		}
		
		else if(isdigit(sequence[i]) || sequence[i]=='.')
		{
			token.TokenType=intT;
			token.TokenName.Empty();
			while(i< seqln && (isdigit(sequence[i]) || sequence[i]=='.'))
			{
				if(sequence[i]=='.') token.TokenType=realT;
				token.TokenName+=sequence[i];
				i++;
			}
			i--;
			if(token.TokenType==intT) token.TokenInt=atoi((LPCSTR)token.TokenName);
			else if(token.TokenType==realT) token.TokenReal=atof((LPCSTR)token.TokenName);

			if(token.TokenType==realT &&token.TokenReal==0.0)
				if(sequence[i]=='.') token.TokenType=dotT;
				else token.TokenType=errorT;
			AddToken(token);
		}
		else if(token.TokenType==nilT)
		{
			for(int j=0;j<SymbolTable.GetSize();j++)
			{
				char buf[100];
				_strnset(buf,0,99);
				strncpy(buf,sequence+i,SymbolTable[j].SymbolString.GetLength());
				_strupr(buf);
				if(strcmp(buf,SymbolTable[j].SymbolString)==0
					&& !IsAllAplhaOrDigit(sequence,i,SymbolTable[j].SymbolString.GetLength()+1) )
				{
					token.TokenType=SymbolTable[j].TokenType;
					i+=SymbolTable[j].SymbolString.GetLength()-1;
					AddToken(token);
					break;
				}
			}
		}
		if(token.TokenType==nilT)
		{
			token.TokenName.Empty();
			
			if(i<seqln && (isalpha(sequence[i]) || isdigit(sequence[i]) ) )
			{
				while(i<seqln  && (isalpha(sequence[i]) || isdigit(sequence[i]) || sequence[i]=='_') )
				{
					token.TokenType=idT;
					token.TokenName+=sequence[i];
					i++;
				}
				if(i<seqln) 
					if(sequence[i]!=' ') 
						i--;
				if(token.TokenType!=nilT) 
					AddToken(token);
			}
		}
		if(token.TokenType==nilT)
		{
			token.TokenType=errorT;
			AddToken(token);
		}
	}
	token.TokenType=endfileT;
	AddToken(token);
	CurrentPosition=TokenSequence.GetHeadPosition();
}

CToken CLexicalAnalyser::GetCurrentToken()
{
	CToken NIL;
	NIL.TokenType=nilT;

	if(CurrentPosition==NULL) 
		return NIL;
	else
		return TokenSequence.GetAt(CurrentPosition);
}

CToken CLexicalAnalyser::NextToken()
{
	CToken NIL;
	NIL.TokenType=nilT;

	if(CurrentPosition==NULL) 
		return NIL;
	else
		return TokenSequence.GetNext(CurrentPosition);
}

void CLexicalAnalyser::AddToken(CToken token)
{
	CurrentPosition=TokenSequence.AddTail(token);
}

BOOL CLexicalAnalyser::IsSequenceEmpty()
{
	return TokenSequence.IsEmpty();
}

bool CLexicalAnalyser::IsAllAplhaOrDigit(CString s, int begin, int length)
{
	if(begin+length>=s.GetLength())
		return false;
	for(int i=begin;i<begin+length;i++)
	{
		if(!isalpha(s[i]) && !isdigit(s[i]) && s[i]!='_') 
			return false;
	}

	return true;
}


void CLexicalAnalyser::ResetPosition()
{
	CurrentPosition=TokenSequence.GetHeadPosition();	
}

void CLexicalAnalyser::ClearAllTokens()
{
	while(!TokenSequence.IsEmpty())
		TokenSequence.RemoveHead();
}

POSITION CLexicalAnalyser::GetCurrentPosition()
{
	return CurrentPosition;
}

void CLexicalAnalyser::SetCurrentPosition(POSITION pos)
{
	CurrentPosition=pos;
}

void CLexicalAnalyser::print()
{
	ResetPosition();
	printf("\n-------------------------------------\n");
	do
	{
		CToken t = GetCurrentToken();
		//****************
		switch(t.getTokenType())
		{
		case nilT: printf("nilT "); break;
		case idT: printf("idT(%s) ",t.getTokenName()); break;
		case intT: printf("intT(%i) ",t.getTokenInt()); break;
		case realT: printf("realT(%f) ",t.getTokenReal()); break;
		case stringT: printf("stringT(%s) ",t.getTokenName()); break;
		case errorT: printf("errorT "); break;
		case endlineT: printf("endlineT "); break;
		case endfileT: printf("endfileT "); break;
		case commaT: printf("commaT "); break;
		case leftpT: printf("leftpT "); break;
		case rightpT: printf("rightpT "); break;
		case ekvT: printf("ekvT "); break;
		case dotT: printf("dotT "); break;
		case asterixT: printf("asterixT "); break;

		case binaryT: printf("binaryT "); break;
		case bitT: printf("bitT "); break;
		case byteT: printf("counterT "); break;
		case counterT: printf("counterT "); break;
		case currencyT: printf("currencyT "); break;
		case datetimeT: printf("datetimeT "); break;
		case doubleT: printf("doubleT "); break;
		case guidT: printf("guidT "); break;
		case longT: printf("longT "); break;
		case longbinaryT: printf("longbinaryT "); break;
		case longtextT: printf("longtextT "); break;
		case shortT: printf("shortT "); break;
		case singleT: printf("singleT "); break;
		case textT: printf("textT "); break;
		case valueT: printf("valueT "); break;

		case addT: printf("addT "); break;
		case allT: printf("allT "); break;
		case alterT: printf("alterT "); break;
		case andT: printf("andT "); break;
		case anyT: printf("anyT "); break;
		case asT: printf("asT "); break;
		case ascT: printf("ascT "); break;
		case avgT: printf("avgT "); break;
		case betweenT: printf("betweenT "); break;
		case byT: printf("byT "); break;
		case columnT: printf("columnT "); break;
		case constraintT: printf("constraintT "); break;
		case countT: printf("countT "); break;
		case createT: printf("createT "); break;
		case databaseT: printf("databaseT "); break;
		case deleteT: printf("deleteT "); break;
		case descT: printf("descT "); break;
		case disallowT: printf("disallowT "); break;
		case distinctT: printf("distinctT "); break;
		case distinctrowT: printf("distinctrowT "); break;
		case dropT: printf("dropT "); break;
		case eqvT: printf("eqvT "); break;
		case existT: printf("existT "); break;
		case foreignT: printf("foreignT "); break;
		case fromT: printf("fromT "); break;
		case groupT: printf("groupT "); break;
		case havingT: printf("havingT "); break;
		case ignoreT: printf("ignoreT "); break;
		case impT: printf("impT "); break;
		case inT: printf("inT "); break;
		case indexT: printf("indexT "); break;
		case innerT: printf("innerT "); break;
		case insertT: printf("insertT "); break;
		case intoT: printf("intoT "); break;
		case isT: printf("isT "); break;
		case joinT: printf("joinT "); break;
		case keyT: printf("keyT "); break;
		case leftT: printf("leftT "); break;
		case likeT: printf("likeT "); break;
		case maxT: printf("maxT "); break;
		case minT: printf("minT "); break;
		case modT: printf("modT "); break;
		case notT: printf("notT "); break;
		case nullT: printf("nullT "); break;
		case onT: printf("onT "); break;
		case optionT: printf("optionT "); break;
		case orT: printf("orT "); break;
		case orderT: printf("orderT "); break;
		case owneraccessT: printf("owneraccessT "); break;
		case parametersT: printf("parametersT "); break;
		case percentT: printf("percentT "); break;
		case pivotT: printf("pivotT "); break;
		case primaryT: printf("primaryT "); break;
		case procedureT: printf("procedureT "); break;
		case referencesT: printf("referencesT "); break;
		case rightT: printf("rightT "); break;
		case selectT: printf("selectT "); break;
		case setT: printf("setT "); break;
		case someT: printf("someT "); break;
		case stdevT: printf("stdevT "); break;
		case stdevpT: printf("stdevpT "); break;
		case sumT: printf("sumT "); break;
		case tableT: printf("tableT "); break;
		case topT: printf("topT "); break;
		case transformT: printf("transformT "); break;
		case unionT: printf("unionT "); break;
		case uniqueT: printf("uniqueT "); break;
		case updateT: printf("updateT "); break;
		case valuesT: printf("valuesT "); break;
		case varT: printf("varT "); break;
		case varpT: printf("varpT "); break;
		case viewT: printf("viewT "); break;
		case whereT: printf("whereT "); break;
		case withT: printf("withT "); break;
		case xorT: printf("xorT "); break;
		}
		//****************
		NextToken();
	} while(CurrentPosition);
	printf("\n\n");
	ResetPosition();
}
